import re

from typing import Any, Dict, List, Union

from sqlalchemy import distinct, func, null, or_
from sqlalchemy.orm import aliased

from conda_store_server import conda_utils, orm, schema, utils


def list_namespaces(db, show_soft_deleted: bool = False):
    filters = []
    if not show_soft_deleted:
        filters.append(orm.Namespace.deleted_on == null())

    return db.query(orm.Namespace).filter(*filters)


def ensure_namespace(db, name: str):
    namespace = get_namespace(db, name=name)
    if namespace is None:
        namespace = create_namespace(db, name=name)
        db.commit()
    return namespace


def get_namespace(
    db, name: str = None, id: int = None, show_soft_deleted: bool = True
) -> orm.Namespace:
    filters = []
    if name:
        filters.append(orm.Namespace.name == name)
    if id:
        filters.append(orm.Namespace.id == id)
    if not show_soft_deleted:
        filters.append(orm.Namespace.deleted_on == null())
    return db.query(orm.Namespace).filter(*filters).first()


def create_namespace(db, name: str):
    if re.fullmatch(f"[{schema.ALLOWED_CHARACTERS}]+", name) is None:
        raise ValueError(
            f"Namespace='{name}' is not valid does not match regex {schema.NAMESPACE_REGEX}"
        )

    namespace = orm.Namespace(name=name)
    db.add(namespace)
    return namespace


def update_namespace(
    db,
    name: str,
    metadata_: Dict[str, Any] = None,
    role_mappings: Dict[str, List[str]] = None,
):
    namespace = get_namespace(db, name)
    if namespace is None:
        raise ValueError(f"Namespace='{name}' not found")

    if metadata_ is not None:
        namespace.metadata_ = metadata_

    if role_mappings is not None:
        # deletes all the existing role mappings ...
        for rm in namespace.role_mappings:
            db.delete(rm)

        # ... before adding all the new ones
        mappings_orm = []
        for entity, roles in role_mappings.items():
            for role in roles:
                mapping_orm = orm.NamespaceRoleMapping(
                    namespace_id=namespace.id,
                    namespace=namespace,
                    entity=entity,
                    role=role,
                )
                mappings_orm.append(mapping_orm)

        namespace.role_mappings = mappings_orm

    db.commit()

    return namespace


# v2 API
def update_namespace_metadata(
    db,
    name: str,
    metadata_: Dict[str, Any] = None,
):
    namespace = get_namespace(db, name)
    if namespace is None:
        raise ValueError(f"Namespace='{name}' not found")

    if metadata_ is not None:
        namespace.metadata_ = metadata_

    return namespace


# v2 API
def get_namespace_roles(
    db,
    name: str,
):
    """Which namespaces can access namespace 'name'?"""
    namespace = get_namespace(db, name)
    if namespace is None:
        raise ValueError(f"Namespace='{name}' not found")

    nrm = aliased(orm.NamespaceRoleMappingV2)
    this = aliased(orm.Namespace)
    other = aliased(orm.Namespace)
    q = (
        db.query(nrm.id, this.name, other.name, nrm.role)
        .filter(nrm.namespace_id == namespace.id)
        .filter(nrm.namespace_id == this.id)
        .filter(nrm.other_namespace_id == other.id)
        .all()
    )
    return [schema.NamespaceRoleMappingV2.from_list(x) for x in q]


# v2 API
def get_other_namespace_roles(
    db,
    name: str,
):
    """To which namespaces does namespace 'name' have access?"""
    namespace = get_namespace(db, name)
    if namespace is None:
        raise ValueError(f"Namespace='{name}' not found")

    nrm = aliased(orm.NamespaceRoleMappingV2)
    this = aliased(orm.Namespace)
    other = aliased(orm.Namespace)
    q = (
        db.query(nrm.id, this.name, other.name, nrm.role)
        .filter(nrm.other_namespace_id == namespace.id)
        .filter(nrm.namespace_id == this.id)
        .filter(nrm.other_namespace_id == other.id)
        .all()
    )
    return [schema.NamespaceRoleMappingV2.from_list(x) for x in q]


# v2 API
def delete_namespace_roles(
    db,
    name: str,
):
    namespace = get_namespace(db, name)
    if namespace is None:
        raise ValueError(f"Namespace='{name}' not found")

    nrm = orm.NamespaceRoleMappingV2
    db.query(nrm).filter(nrm.namespace_id == namespace.id).delete()


# v2 API
def get_namespace_role(
    db,
    name: str,
    other: str,
):
    namespace = get_namespace(db, name)
    if namespace is None:
        raise ValueError(f"Namespace='{name}' not found")

    other_namespace = get_namespace(db, other)
    if other_namespace is None:
        raise ValueError(f"Namespace='{other}' not found")

    nrm = aliased(orm.NamespaceRoleMappingV2)
    this = aliased(orm.Namespace)
    other = aliased(orm.Namespace)
    q = (
        db.query(nrm.id, this.name, other.name, nrm.role)
        .filter(nrm.namespace_id == namespace.id)
        .filter(nrm.other_namespace_id == other_namespace.id)
        .filter(nrm.namespace_id == this.id)
        .filter(nrm.other_namespace_id == other.id)
        .first()
    )
    if q is None:
        return None
    return schema.NamespaceRoleMappingV2.from_list(q)


# v2 API
def create_namespace_role(
    db,
    name: str,
    other: str,
    role: str,
):
    namespace = get_namespace(db, name)
    if namespace is None:
        raise ValueError(f"Namespace='{name}' not found")

    other_namespace = get_namespace(db, other)
    if other_namespace is None:
        raise ValueError(f"Namespace='{other}' not found")

    db.add(
        orm.NamespaceRoleMappingV2(
            namespace_id=namespace.id,
            other_namespace_id=other_namespace.id,
            role=role,
        )
    )


# v2 API
def update_namespace_role(
    db,
    name: str,
    other: str,
    role: str,
):
    namespace = get_namespace(db, name)
    if namespace is None:
        raise ValueError(f"Namespace='{name}' not found")

    other_namespace = get_namespace(db, other)
    if other_namespace is None:
        raise ValueError(f"Namespace='{other}' not found")

    nrm = orm.NamespaceRoleMappingV2
    q = (
        db.query(nrm)
        .filter(nrm.namespace_id == namespace.id)
        .filter(nrm.other_namespace_id == other_namespace.id)
        .first()
    )
    # Important: this modifies a field of a table entry and calls 'db.add'
    # instead of using '.update({"role": role})' on the query because the latter
    # would bypass the ORM validation logic, which maps the 'editor' role to
    # 'developer'
    q.role = role
    db.add(q)


# v2 API
def delete_namespace_role(
    db,
    name: str,
    other: str,
):
    namespace = get_namespace(db, name)
    if namespace is None:
        raise ValueError(f"Namespace='{name}' not found")

    other_namespace = get_namespace(db, other)
    if other_namespace is None:
        raise ValueError(f"Namespace='{other}' not found")

    nrm = orm.NamespaceRoleMappingV2
    db.query(nrm).filter(nrm.namespace_id == namespace.id).filter(
        nrm.other_namespace_id == other_namespace.id
    ).delete()


def delete_namespace(db, name: str = None, id: int = None):
    namespace = get_namespace(db, name=name, id=id)
    if namespace:
        db.delete(namespace)


def list_environments(
    db,
    namespace: str = None,
    name: str = None,
    status: schema.BuildStatus = None,
    packages: List[str] = None,
    artifact: schema.BuildArtifactType = None,
    search: str = None,
    show_soft_deleted: bool = False,
):
    query = db.query(orm.Environment).join(orm.Environment.namespace)

    if namespace:
        query = query.filter(orm.Namespace.name == namespace)

    if name:
        query = query.filter(orm.Environment.name == name)

    if search:
        query = query.filter(
            or_(
                orm.Namespace.name.contains(search, autoescape=True),
                orm.Environment.name.contains(search, autoescape=True),
            )
        )

    if not show_soft_deleted:
        query = query.filter(orm.Environment.deleted_on == null())

    if status or artifact or packages:
        query = query.join(orm.Environment.current_build)

    if status:
        query = query.filter(orm.Build.status == status)

    if artifact:
        # DOCKER_BLOB can return multiple results
        # use DOCKER_MANIFEST instead
        if artifact == schema.BuildArtifactType.DOCKER_BLOB:
            artifact = schema.BuildArtifactType.DOCKER_MANIFEST
        query = query.join(orm.Build.build_artifacts).filter(
            orm.BuildArtifact.artifact_type == artifact
        )

    if packages:
        query = (
            query.join(orm.Build.package_builds)
            .join(orm.CondaPackageBuild.package)
            .filter(orm.CondaPackage.name.in_(packages))
            .group_by(orm.Namespace.name, orm.Environment.name, orm.Environment.id)
            .having(func.count() == len(packages))
        )

    return query


def ensure_environment(
    db,
    name: str,
    namespace_id: int,
    description: str = None,
):
    environment = get_environment(db, name=name, namespace_id=namespace_id)
    if environment is None:
        environment = create_environment(
            db, name=name, namespace_id=namespace_id, description=description
        )
        db.commit()
    elif description:
        update_environment(
            db, name=name, namespace_id=namespace_id, description=description
        )
        db.commit()
    return environment


def create_environment(db, name: str, namespace_id: int, description: str):
    environment = orm.Environment(
        name=name, namespace_id=namespace_id, description=description
    )
    db.add(environment)
    return environment


def update_environment(db, name: str, namespace_id: int, description: str):
    environment = get_environment(db, name=name, namespace_id=namespace_id)
    environment.description = description
    return environment


def get_environment(
    db,
    name: str = None,
    namespace: str = None,
    namespace_id: int = None,
    id: int = None,
):
    filters = []
    if namespace:
        filters.append(orm.Namespace.name == namespace)
    if namespace_id:
        filters.append(orm.Namespace.id == namespace_id)
    if name:
        filters.append(orm.Environment.name == name)
    if id:
        filters.append(orm.Environment.id == id)

    return db.query(orm.Environment).join(orm.Namespace).filter(*filters).first()


def ensure_specification(
    db,
    specification: Union[schema.CondaSpecification, schema.LockfileSpecification],
    is_lockfile: bool = False,
):
    specification_sha256 = utils.datastructure_hash(specification.dict())
    specification_orm = get_specification(db, sha256=specification_sha256)

    if specification_orm is None:
        specification_orm = create_speficication(
            db, specification, is_lockfile=is_lockfile
        )
        db.commit()

    return specification_orm


def create_speficication(
    db,
    specification: Union[schema.CondaSpecification, schema.LockfileSpecification],
    is_lockfile: bool = False,
):
    specification_orm = orm.Specification(specification.dict(), is_lockfile=is_lockfile)
    db.add(specification_orm)
    return specification_orm


def list_specifications(db, search=None):
    filters = []
    return db.query(orm.Specification).filter(*filters)


def get_specification(db, sha256: str):
    filters = [orm.Specification.sha256 == sha256]
    return db.query(orm.Specification).filter(*filters).first()


def create_solve(db, specification_id: int):
    solve = orm.Solve(specification_id=specification_id)
    db.add(solve)
    return solve


def get_solve(db, solve_id: int):
    return db.query(orm.Solve).filter(orm.Solve.id == solve_id).first()


def list_solves(db):
    return db.query(orm.Solve)


def list_builds(
    db,
    status: schema.BuildStatus = None,
    packages: List[str] = None,
    artifact: schema.BuildArtifactType = None,
    environment_id: str = None,
    name: str = None,
    namespace: str = None,
    show_soft_deleted: bool = False,
):
    query = (
        db.query(orm.Build).join(orm.Build.environment).join(orm.Environment.namespace)
    )

    if status:
        query = query.filter(orm.Build.status == status)

    if environment_id:
        query = query.filter(orm.Build.environment_id == environment_id)

    if name:
        query = query.filter(orm.Environment.name == name)

    if namespace:
        query = query.filter(orm.Namespace.name == namespace)

    if not show_soft_deleted:
        query = query.filter(orm.Build.deleted_on == null())

    if artifact:
        # DOCKER_BLOB can return multiple results
        # use DOCKER_MANIFEST instead
        if artifact == schema.BuildArtifactType.DOCKER_BLOB:
            artifact = schema.BuildArtifactType.DOCKER_MANIFEST
        query = query.join(orm.Build.build_artifacts).filter(
            orm.BuildArtifact.artifact_type == artifact
        )

    if packages:
        query = (
            query.join(orm.Build.packages)
            .filter(orm.CondaPackage.name.in_(packages))
            .group_by(orm.Build.id)
            .having(func.count() == len(packages))
        )

    return query


def create_build(db, environment_id: int, specification_id: int):
    build = orm.Build(environment_id=environment_id, specification_id=specification_id)
    db.add(build)
    return build


def get_build(db, build_id: int):
    return db.query(orm.Build).filter(orm.Build.id == build_id).first()


def get_build_packages(
    db, build_id: int, search: str = None, exact: bool = False, build: str = None
):
    filters = [(orm.Build.id == build_id)]
    if search:
        if exact:
            filters.append(orm.CondaPackage.name.like(search.replace("%", r"\%")))
        else:
            filters.append(orm.CondaPackage.name.contains(search, autoescape=True))
    if build:
        filters.append(orm.CondaPackageBuild.build.contains(build, autoescape=True))

    return (
        db.query(orm.CondaPackage)
        .join(orm.Build.package_builds)
        .join(orm.CondaPackageBuild.package)
        .join(orm.CondaPackage.channel)
        .filter(*filters)
    )


def get_build_lockfile_legacy(db, build_id: int):
    build = db.query(orm.Build).filter(orm.Build.id == build_id).first()
    packages = [
        f"{row.package.channel.name}/{row.subdir}/{row.package.name}-{row.package.version}-{row.build}.tar.bz2#{row.md5}"
        for row in build.package_builds
    ]
    return """#platform: {0}
@EXPLICIT
{1}
""".format(
        conda_utils.conda_platform(), "\n".join(packages)
    )


def get_build_artifact_types(db, build_id: int):
    return (
        db.query(orm.BuildArtifact.artifact_type)
        .filter(orm.BuildArtifact.build_id == build_id)
        .distinct()
    )


def list_build_artifacts(
    db,
    build_id: int = None,
    key: str = None,
    excluded_artifact_types: List[schema.BuildArtifactType] = None,
    included_artifact_types: List[schema.BuildArtifactType] = None,
):
    filters = []
    if build_id:
        filters.append(orm.BuildArtifact.build_id == build_id)
    if key:
        filters.append(orm.BuildArtifact.key == key)
    if excluded_artifact_types:
        filters.append(
            func.not_(orm.BuildArtifact.artifact_type.in_(excluded_artifact_types))
        )
    if included_artifact_types:
        filters.append(orm.BuildArtifact.artifact_type.in_(included_artifact_types))

    return db.query(orm.BuildArtifact).filter(*filters)


def get_build_artifact(db, build_id: int, key: str):
    return (
        db.query(orm.BuildArtifact)
        .filter(orm.BuildArtifact.build_id == build_id, orm.BuildArtifact.key == key)
        .first()
    )


def ensure_conda_channel(db, channel_name: str):
    conda_channel = get_conda_channel(db, channel_name)
    if conda_channel is None:
        conda_channel = create_conda_channel(db, channel_name)
        db.commit()
    return conda_channel


def list_conda_channels(db):
    filters = []
    return db.query(orm.CondaChannel).filter(*filters)


def create_conda_channel(db, channel_name: str):
    channel = orm.CondaChannel(name=channel_name, last_update=None)
    db.add(channel)
    return channel


def get_conda_channel(db, channel_name: str):
    return (
        db.query(orm.CondaChannel).filter(orm.CondaChannel.name == channel_name).first()
    )


def get_conda_package(db, channel_id: int, name: str, version: str):
    return (
        db.query(orm.CondaPackage)
        .filter(orm.CondaPackage.channel_id == channel_id)
        .filter(orm.CondaPackage.name == name)
        .filter(orm.CondaPackage.version == version)
        .first()
    )


def get_conda_package_build(db, package_id: int, subdir: str, build: str):
    return (
        db.query(orm.CondaPackageBuild)
        .filter(orm.CondaPackageBuild.package_id == package_id)
        .filter(orm.CondaPackageBuild.subdir == subdir)
        .filter(orm.CondaPackageBuild.build == build)
        .first()
    )


def create_or_ignore_conda_package(db, package_record: Dict):
    # first create the conda channel
    channel = package_record["channel_id"]
    if channel == "https://conda.anaconda.org/pypi":
        # ignore pypi package for now
        return None

    channel_orm = get_conda_channel(db, channel)
    if channel_orm is None:
        channel_orm = create_conda_channel(db, channel)
        db.commit()

    package_record["channel_id"] = channel_orm.id

    # Retrieve the package from the DB if it already exists
    conda_package = get_conda_package(
        db,
        channel_id=package_record["channel_id"],
        name=package_record["name"],
        version=package_record["version"],
    )

    # If it doesn't exist, let's create it in DB
    if conda_package is None:
        conda_package = create_conda_package(db, package_record=package_record)

    # Retrieve the build for this pacakge, if it already exists
    conda_package_build = get_conda_package_build(
        db,
        package_id=conda_package.id,
        subdir=package_record["subdir"],
        build=package_record["build"],
    )

    # If it doesn't exist, let's create it in DB
    if conda_package_build is None:
        conda_package_build = create_conda_package_build(
            db,
            package_id=conda_package.id,
            package_record=package_record,
        )

    return conda_package_build


def create_conda_package(db, package_record: Dict):
    conda_package_keys = [
        "channel_id",
        "license",
        "license_family",
        "name",
        "version",
        "summary",
        "description",
    ]

    conda_package = orm.CondaPackage(
        **{k: package_record[k] for k in conda_package_keys}
    )
    db.add(conda_package)
    return conda_package


def create_conda_package_build(db, package_id: int, package_record: Dict):
    conda_package_build_keys = [
        "build",
        "build_number",
        "constrains",
        "depends",
        "md5",
        "sha256",
        "size",
        "subdir",
        "timestamp",
    ]
    conda_package_build = orm.CondaPackageBuild(
        package_id=package_id,
        **{k: package_record[k] for k in conda_package_build_keys},
    )
    db.add(conda_package_build)
    return conda_package_build


def list_conda_packages(db, search: str = None, exact: bool = False, build: str = None):
    filters = []
    if search:
        if exact:
            filters.append(orm.CondaPackage.name.like(search.replace("%", r"\%")))
        else:
            filters.append(orm.CondaPackage.name.contains(search, autoescape=True))
    if build:
        filters.append(orm.CondaPackage.build.contains(build, autoescape=True))

    return db.query(orm.CondaPackage).join(orm.CondaChannel).filter(*filters)


def get_metrics(db):
    metrics = (
        db.query(
            orm.CondaStoreConfiguration.free_storage.label("disk_free"),
            orm.CondaStoreConfiguration.total_storage.label("disk_total"),
            orm.CondaStoreConfiguration.disk_usage,
        )
        .first()
        ._asdict()
    )

    query = db.query(func.count(orm.Build.status), orm.Build.status).group_by(
        orm.Build.status
    )
    for build_count, build_status_enum in query.all():
        metrics[f"build_{build_status_enum.value.lower()}"] = build_count

    metrics["environments"] = db.query(orm.Environment).count()
    return metrics


def get_system_metrics(db):
    return db.query(
        orm.CondaStoreConfiguration.free_storage.label("disk_free"),
        orm.CondaStoreConfiguration.total_storage.label("disk_total"),
        orm.CondaStoreConfiguration.disk_usage,
    ).first()


def get_namespace_metrics(db):
    return (
        db.query(
            orm.Namespace.name,
            func.count(distinct(orm.Environment.id)),
            func.count(distinct(orm.Build.id)),
            func.sum(orm.Build.size),
        )
        .join(orm.Build.environment)
        .join(orm.Environment.namespace)
        .group_by(orm.Namespace.name)
    )


def get_kvstore_key_values(db, prefix: str):
    """Get effective key, values for a particular prefix"""
    return {
        _.key: _.value
        for _ in db.query(orm.KeyValueStore)
        .filter(orm.KeyValueStore.prefix == prefix)
        .all()
    }


def set_kvstore_key_values(db, prefix: str, d: Dict[str, Any], update: bool = True):
    """Set key, values for a particular prefix"""
    for key, value in d.items():
        record = (
            db.query(orm.KeyValueStore)
            .filter(orm.KeyValueStore.prefix == prefix, orm.KeyValueStore.key == key)
            .first()
        )

        if record is None:
            record = orm.KeyValueStore(
                prefix=prefix,
                key=key,
                value=value,
            )
            db.add(record)
            db.commit()
        elif update:
            record.value = value
            db.commit()



import datetime
import hashlib
import typing

import platformdirs


__version__ = "2024.6.1"


CONDA_STORE_DIR = platformdirs.user_data_path(appname="conda-store")


class BuildKey:
    """
    Used to configure the build key format, which identifies a particular
    environment build

    Avoids a cyclic dependency between the `orm` module and the module defining
    `CondaStore.build_key_version`. Because the `orm` module is loaded early on
    startup, we want to delay initialization of the `Build.build_key_version`
    field until `CondaStore.build_key_version` has been read from the config.

    Because the build key version needs to be the same for the entire
    application, this class implements the singleton pattern. Users are expected
    to use class methods instead of creating class instances. All implementation
    details are hidden within the class, preventing potential issues caused by
    cyclic imports.
    """

    # Default version, must be None here. Initialized in CondaStore.build_key_version
    _current_version = None

    _version2_hash_size = 8

    _version3_experimental_hash_size = 32

    def _version1_fmt(build: "Build") -> str:  # noqa: F821
        datetime_format = "%Y%m%d-%H%M%S-%f"
        hash = build.specification.sha256
        timestamp = build.scheduled_on.strftime(datetime_format)
        id = build.id
        name = build.specification.name
        return f"{hash}-{timestamp}-{id}-{name}"

    def _version2_fmt(build: "Build") -> str:  # noqa: F821
        tzinfo = datetime.timezone.utc
        hash = build.specification.sha256[: BuildKey._version2_hash_size]
        timestamp = int(build.scheduled_on.replace(tzinfo=tzinfo).timestamp())
        id = build.id
        name = build.specification.name
        return f"{hash}-{timestamp}-{id}-{name}"

    # Warning: this is an experimental version and can be changed at any time
    def _version3_experimental_fmt(build: "Build") -> str:  # noqa: F821
        # Caches the hash value for faster lookup later
        if build.hash is not None:
            return build.hash

        # Adds namespace here to separate builds performed by different users
        # since all builds are stored in the same directory in v3, see
        # Build.build_path in orm.py. Additionally, this also hashes the
        # timestamp and build id just to make collisions very unlikely
        namespace_name = build.environment.namespace.name
        specification_hash = build.specification.sha256
        tzinfo = datetime.timezone.utc
        timestamp = int(build.scheduled_on.replace(tzinfo=tzinfo).timestamp())
        build_id = build.id
        hash_input = (
            namespace_name + specification_hash + str(timestamp) + str(build_id)
        )
        hash = hashlib.sha256(hash_input.encode("utf-8")).hexdigest()
        hash = hash[: BuildKey._version3_experimental_hash_size]
        build.hash = hash
        return hash

    # version -> fmt function
    _fmt = {
        1: _version1_fmt,
        2: _version2_fmt,
        3: _version3_experimental_fmt,
    }

    @classmethod
    def _check_version(cls, build_key_version: int):
        if build_key_version not in cls.versions():
            raise ValueError(
                f"invalid build key version: {build_key_version}, "
                f"expected: {cls.versions()}"
            )

    @classmethod
    def set_current_version(cls, build_key_version: int) -> int:
        """Sets provided build key version as current and returns it"""
        cls._check_version(build_key_version)
        cls._current_version = build_key_version
        return build_key_version

    @classmethod
    def current_version(cls) -> int:
        """Returns currently selected build key version"""
        # None means the value is not set, likely due to an import error
        assert cls._current_version is not None
        return cls._current_version

    @classmethod
    def versions(cls) -> typing.Tuple[int]:
        """Returns available build key versions"""
        return tuple(cls._fmt.keys())

    @classmethod
    def get_build_key(cls, build: "Build") -> str:  # noqa: F821
        """Returns build key for this build"""
        cls._check_version(build.build_key_version)
        return cls._fmt.get(build.build_key_version)(build)

    @classmethod
    def parse_build_key(
        cls, conda_store: "CondaStore", build_key: str  # noqa: F821
    ) -> int:
        """Returns build id from build key"""
        # This import is here to avoid cyclic imports
        from conda_store_server import orm

        parts = build_key.split("-")
        # Note: cannot rely on the number of dashes to differentiate between
        # versions because name can contain dashes. Instead, this relies on the
        # hash size to infer the format. The name is the last field, so indexing
        # to find the id is okay.
        if build_key[cls._version2_hash_size] == "-":  # v2
            return int(parts[2])  # build_id
        elif "-" not in build_key:  # v3
            with conda_store.get_db() as db:
                build = db.query(orm.Build).filter(orm.Build.hash == build_key).first()
                if build is not None:
                    return build.id
                return None
        else:  # v1
            return int(parts[4])  # build_id



"""add build_key_version

Revision ID: 30b37e725c32
Revises: d78e9889566a
Create Date: 2023-11-17 14:34:40.688865

"""

import sqlalchemy as sa

from alembic import op


# revision identifiers, used by Alembic.
revision = "30b37e725c32"
down_revision = "d78e9889566a"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "build",
        sa.Column(
            "build_key_version", sa.Integer(), nullable=False, server_default="1"
        ),
    )


def downgrade():
    op.drop_column("build", "build_key_version")



from conda_store_server.action.base import action  # noqa

from conda_store_server.action.generate_lockfile import action_solve_lockfile  # noqa
from conda_store_server.action.generate_lockfile import action_save_lockfile  # noqa
from conda_store_server.action.download_packages import (
    action_fetch_and_extract_conda_packages,  # noqa
)
from conda_store_server.action.install_lockfile import action_install_lockfile  # noqa
from conda_store_server.action.install_specification import (
    action_install_specification,  # noqa
)
from conda_store_server.action.generate_conda_export import (
    action_generate_conda_export,  # noqa
)
from conda_store_server.action.generate_conda_pack import (
    action_generate_conda_pack,  # noqa
)
from conda_store_server.action.generate_conda_docker import (
    action_generate_conda_docker,  # noqa
)
from conda_store_server.action.remove_conda_prefix import (
    action_remove_conda_prefix,  # noqa
)
from conda_store_server.action.set_conda_prefix_permissions import (
    action_set_conda_prefix_permissions,  # noqa
)
from conda_store_server.action.get_conda_prefix_stats import (
    action_get_conda_prefix_stats,  # noqa
)
from conda_store_server.action.add_conda_prefix_packages import (
    action_add_conda_prefix_packages,  # noqa
)
from conda_store_server.action.add_lockfile_packages import (
    action_add_lockfile_packages,  # noqa
)
from conda_store_server.action.generate_constructor_installer import (
    action_generate_constructor_installer,  # noqa
)



import datetime
import os
import shutil
import sys
import typing

import yaml

from celery import Task, platforms, shared_task
from celery.execute import send_task
from celery.signals import worker_ready
from filelock import FileLock
from sqlalchemy.orm import Session

from conda_store_server import api, environment, schema, utils
from conda_store_server.build import (
    build_cleanup,
    build_conda_docker,
    build_conda_env_export,
    build_conda_environment,
    build_conda_pack,
    build_constructor_installer,
    solve_conda_environment,
)
from conda_store_server.worker.app import CondaStoreWorker


@worker_ready.connect
def at_start(sender, **k):
    with sender.app.connection():
        sender.app.send_task("task_initialize_worker")
        sender.app.send_task("task_update_conda_channels")
        sender.app.send_task("task_watch_paths")
        sender.app.send_task("task_cleanup_builds")


class WorkerTask(Task):
    @property
    def worker(self):
        if not hasattr(self, "_worker"):
            self._worker = CondaStoreWorker()

            # hook to allow for traitlet configuration via celery config
            if "traitlets" in self.app.conf:
                from traitlets.config.loader import Config

                config = Config(**self.app.conf.traitlets)
                self._worker.update_config(config)

            self._worker.initialize()

        # Installs a signal handler that terminates the process on Ctrl-C
        # (SIGINT)
        # Note: the following would not be enough to terminate beat and other
        # tasks, which is why the code below explicitly calls sys.exit
        # > from celery.apps.worker import install_worker_term_hard_handler
        # > install_worker_term_hard_handler(self._worker, sig='SIGINT')
        def _shutdown(*args, **kwargs):
            return sys.exit(1)

        platforms.signals["INT"] = _shutdown

        return self._worker


# Signals to the server that the worker is running, see _check_worker in
# CondaStoreServer
@shared_task(base=WorkerTask, name="task_initialize_worker", bind=True)
def task_initialize_worker(self):
    from conda_store_server import orm

    conda_store = self.worker.conda_store

    with conda_store.session_factory() as db:
        db.add(orm.Worker(initialized=True))
        db.commit()


@shared_task(base=WorkerTask, name="task_watch_paths", bind=True)
def task_watch_paths(self):
    conda_store = self.worker.conda_store

    with conda_store.session_factory() as db:
        settings = conda_store.get_settings(db)

        conda_store.configuration(db).update_storage_metrics(
            db, conda_store.store_directory
        )

        environment_paths = environment.discover_environments(self.worker.watch_paths)
        for path in environment_paths:
            with open(path) as f:
                conda_store.register_environment(
                    db,
                    specification=yaml.safe_load(f),
                    namespace=settings.filesystem_namespace,
                    force=False,
                )


@shared_task(base=WorkerTask, name="task_update_storage_metrics", bind=True)
def task_update_storage_metrics(self):
    conda_store = self.worker.conda_store
    with conda_store.session_factory() as db:
        conda_store.configuration(db).update_storage_metrics(
            db, conda_store.store_directory
        )


@shared_task(base=WorkerTask, name="task_cleanup_builds", bind=True)
def task_cleanup_builds(
    self,
    build_ids: typing.List[str] = None,
    reason: str = None,
    is_canceled: bool = False,
):
    conda_store = self.worker.conda_store
    with conda_store.session_factory() as db:
        build_cleanup(db, conda_store, build_ids, reason, is_canceled)


"""
Pierre - May 29th 2022
This is a different version of task_update_conda_channels.
It's designed to run one task per channel, with a lock to avoid triggering twice a task for a given channel.
Two scenarios :
- either you're using Redis (that should be the case if you have multiple celery workers) : then the lock is handled by Redis.
- either you don't (that's fine if you're using a single celery worker) : then the lock is a Filelock

The reason behind having the task running only once at a time is
to avoid integrity exceptions when running channel.update_packages.

Sources :
Lock : http://loose-bits.com/2010/10/distributed-task-locking-in-celery.html
Redis : https://pypi.org/project/redis/
https://stackoverflow.com/questions/12003221/celery-task-schedule-ensuring-a-task-is-only-executed-one-at-a-time

"""


@shared_task(base=WorkerTask, name="task_update_conda_channels", bind=True)
def task_update_conda_channels(self):
    conda_store = self.worker.conda_store
    with conda_store.session_factory() as db:
        conda_store.ensure_conda_channels(db)
        for channel in api.list_conda_channels(db):
            send_task("task_update_conda_channel", args=[channel.name], kwargs={})


@shared_task(base=WorkerTask, name="task_update_conda_channel", bind=True)
def task_update_conda_channel(self, channel_name):
    conda_store = self.worker.conda_store
    with conda_store.session_factory() as db:
        settings = conda_store.get_settings(db)

        # sanitize the channel name as it's an URL, and it's used for the lock.
        sanitizing = {
            "https": "",
            "http": "",
            ":": "",
            "/": "_",
            "?": "",
            "&": "_",
            "=": "_",
        }
        channel_name_sanitized = channel_name
        for k, v in sanitizing.items():
            channel_name_sanitized = channel_name_sanitized.replace(k, v)

        task_key = f"lock_{self.name}_{channel_name_sanitized}"

        is_locked = False

        if conda_store.redis_url is not None:
            lock = conda_store.redis.lock(task_key, timeout=60 * 15)  # timeout 15min
        else:
            lockfile_path = os.path.join(f"/tmp/task_lock_{task_key}")
            lock = FileLock(lockfile_path, timeout=60 * 15)

        try:
            is_locked = lock.acquire(blocking=False)

            if is_locked:
                channel = api.get_conda_channel(db, channel_name)

                conda_store.log.debug(f"updating packages for channel {channel.name}")
                channel.update_packages(db, subdirs=settings.conda_platforms)

            else:
                conda_store.log.debug(
                    f"skipping updating packages for channel {channel_name} - already in progress"
                )

        except TimeoutError:
            if conda_store.redis_url is None:
                conda_store.log.warning(
                    f"Timeout when acquiring lock with key {task_key} - We assume the task is already being run"
                )
                is_locked = False

        finally:
            if is_locked:
                lock.release()


@shared_task(base=WorkerTask, name="task_solve_conda_environment", bind=True)
def task_solve_conda_environment(self, solve_id):
    conda_store = self.worker.conda_store

    with conda_store.session_factory() as db:
        solve = api.get_solve(db, solve_id)
        solve_conda_environment(db, conda_store, solve)


@shared_task(base=WorkerTask, name="task_build_conda_environment", bind=True)
def task_build_conda_environment(self, build_id):
    conda_store = self.worker.conda_store

    with conda_store.session_factory() as db:
        build = api.get_build(db, build_id)
        build_conda_environment(db, conda_store, build)


@shared_task(base=WorkerTask, name="task_build_conda_env_export", bind=True)
def task_build_conda_env_export(self, build_id):
    conda_store = self.worker.conda_store
    with conda_store.session_factory() as db:
        build = api.get_build(db, build_id)
        build_conda_env_export(db, conda_store, build)


@shared_task(base=WorkerTask, name="task_build_conda_pack", bind=True)
def task_build_conda_pack(self, build_id):
    conda_store = self.worker.conda_store
    with conda_store.session_factory() as db:
        build = api.get_build(db, build_id)
        build_conda_pack(db, conda_store, build)


@shared_task(base=WorkerTask, name="task_build_conda_docker", bind=True)
def task_build_conda_docker(self, build_id):
    conda_store = self.worker.conda_store
    with conda_store.session_factory() as db:
        build = api.get_build(db, build_id)
        build_conda_docker(db, conda_store, build)


@shared_task(base=WorkerTask, name="task_build_constructor_installer", bind=True)
def task_build_constructor_installer(self, build_id):
    conda_store = self.worker.conda_store
    with conda_store.session_factory() as db:
        build = api.get_build(db, build_id)
        build_constructor_installer(db, conda_store, build)


@shared_task(base=WorkerTask, name="task_update_environment_build", bind=True)
def task_update_environment_build(self, environment_id):
    conda_store = self.worker.conda_store
    with conda_store.session_factory() as db:
        environment = api.get_environment(db, id=environment_id)

        conda_prefix = environment.current_build.build_path(conda_store)
        environment_prefix = environment.current_build.environment_path(conda_store)

        if environment_prefix is not None:
            utils.symlink(conda_prefix, environment_prefix)

        if conda_store.post_update_environment_build_hook:
            conda_store.post_update_environment_build_hook(conda_store, environment)


def delete_build_artifact(db: Session, conda_store, build_artifact):
    if build_artifact.artifact_type == schema.BuildArtifactType.DIRECTORY:
        # ignore key
        conda_prefix = build_artifact.build.build_path(conda_store)
        # be REALLY sure this is a directory within store directory
        if str(conda_prefix).startswith(conda_store.store_directory) and os.path.isdir(
            conda_prefix
        ):
            shutil.rmtree(conda_prefix)
            db.delete(build_artifact)
    elif build_artifact.artifact_type == schema.BuildArtifactType.CONTAINER_REGISTRY:
        pass
        # # container registry tag deletion is not generally implemented
        # # the underlying library `python_docker` is already capable
        # conda_store.container_registry.delete_image(build_artifact.key)
    else:
        conda_store.log.info(f"deleting {build_artifact.key}")
        conda_store.storage.delete(db, build_artifact.build.id, build_artifact.key)


@shared_task(base=WorkerTask, name="task_delete_build", bind=True)
def task_delete_build(self, build_id):
    conda_store = self.worker.conda_store
    with conda_store.session_factory() as db:
        settings = conda_store.get_settings(db)

        build = api.get_build(db, build_id)

        # Deletes build artifacts for this build
        conda_store.log.info(f"deleting artifacts for build={build.id}")
        for build_artifact in api.list_build_artifacts(
            db,
            build_id=build_id,
            excluded_artifact_types=settings.build_artifacts_kept_on_deletion,
        ).all():
            delete_build_artifact(db, conda_store, build_artifact)

        # Updates build size and marks build as deleted
        build.deleted_on = datetime.datetime.utcnow()
        build.size = 0

        db.commit()


@shared_task(base=WorkerTask, name="task_delete_environment", bind=True)
def task_delete_environment(self, environment_id):
    conda_store = self.worker.conda_store
    with conda_store.session_factory() as db:
        environment = api.get_environment(db, id=environment_id)

        for build in environment.builds:
            conda_store.log.info(f"deleting artifacts for build={build.id}")
            for build_artifact in api.list_build_artifacts(
                db,
                build_id=build.id,
            ).all():
                delete_build_artifact(db, conda_store, build_artifact)

        db.delete(environment)
        db.commit()


@shared_task(base=WorkerTask, name="task_delete_namespace", bind=True)
def task_delete_namespace(self, namespace_id):
    conda_store = self.worker.conda_store
    with conda_store.session_factory() as db:
        namespace = api.get_namespace(db, id=namespace_id)

        for environment_orm in namespace.environments:
            for build in environment_orm.builds:
                conda_store.log.info(f"deleting artifacts for build={build.id}")
                for build_artifact in api.list_build_artifacts(
                    db,
                    build_id=build.id,
                ).all():
                    delete_build_artifact(db, conda_store, build_artifact)
            db.delete(environment_orm)
        db.delete(namespace)
        db.commit()



import pathlib

import conda_pack

from conda_store_server import action


@action.action
def action_generate_conda_pack(
    context,
    conda_prefix: pathlib.Path,
    output_filename: pathlib.Path,
):
    conda_pack.pack(
        prefix=str(conda_prefix),
        output=str(output_filename),
        ignore_missing_files=True,
    )
